"""A linter for alidist recipes.

Combines shellcheck for recipes with yamllint and some internal checks for a
recipe's YAML header.
"""

# alidistlint._version is generated by setuptools_scm.
from alidistlint._version import version

__version__ = version
